use aoc2023_niels_overkamp::common::{self, AOCResult};
use std::io::Write;

const DAY: &str = "day1";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

fn to_digit(s: String) -> u32 {
    if s.len() == 1 {
        s.chars().next().expect("is not a digit").to_digit(10).unwrap()
    } else {
        match s.as_str() {
            "one" | "eno" => 1,
            "two" | "owt" => 2,
            "three" | "eerht" => 3,
            "four" | "ruof" => 4,
            "five" | "evif" => 5,
            "six" | "xis" => 6,
            "seven" | "neves" => 7,
            "eight" | "thgie" => 8,
            "nine" | "enin" => 9,
            s => panic!("{} is not a digit", s)
        }
    }
}

pub fn run(input: &Vec<String>) -> AOCResult {
    let mut sum1: u32 = 0;
    for line in input {
        let mut lower = None;
        let mut upper = None;
        for c in line.chars() {
            if let Some(d) = c.to_digit(10) {
                upper.get_or_insert(d);
                lower = Some(d);
            }
        }
        if let (Some(upper), Some(lower)) = (upper, lower) {
            sum1 += upper * 10 + lower;
        }
    }
    
    let mut sum2: u32 = 0;
    let re_reg = regex::Regex::new(r"(one|two|three|four|five|six|seven|eight|nine|\d)").unwrap();
    let re_rev = regex::Regex::new(r"(eno|owt|eerht|ruof|evif|xis|neves|thgie|enin|\d)").unwrap();    
    for line in input {        
        if let Some(capture) = re_reg.captures(line) {
            let upper = to_digit(capture.get(1).map(|m| m.as_str().to_owned()).ok_or::<String>(format!("{} did not capture the first digit", &line).into())?);
            if let Some(capture) = re_rev.captures(&line.chars().rev().collect::<String>()) {
                let lower = to_digit(capture.get(1).map(|m| m.as_str().to_owned()).ok_or::<String>(format!("{} did not capture the second digit", &line).into())?);
                sum2 += upper * 10 + lower;
            } else {
                return Err(format!("{} did not capture the second digit", line).into());
            }
        } else {
            return Err(format!("{} did not capture the first digit", line).into());
        }
    }
    
    Ok([Some(sum1.to_string()), Some(sum2.to_string())])
}

#[test]
pub fn test_day1() {
    assert!(common::run_test(DAY, &run))
}
