use aoc2023_niels_overkamp::common::{self, AOCResult};
use std::iter;

const DAY: &str = "day13";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

fn find_mirror_plane(horizontal: &Vec<u16>, vertical: &Vec<u16>) -> Option<usize> {
    let cases = [(horizontal, 100, false), (horizontal, 100, true), (vertical, 1, false), (vertical, 1, true)];
    for (grid, mult, invert) in cases.into_iter() {
        let l = grid.len() - 1;
        let mut index = 1;
        while index < grid.len() {
            let mut mirrors = true;
            for offset in 0..=(index/2) {
                if (invert && grid[l-offset] != grid[l-index+offset]) 
                    || (!invert && grid[offset] != grid[index-offset]) 
                {
                    mirrors = false;
                    break;
                }
            }
            if mirrors {
                if invert {
                    return Some(mult * (l - (index / 2)));
                } else {
                    return Some(mult * ((index / 2) + 1));
                }
            }
            index += 2;
        }
    }
    return None;
}

fn find_near_mirror_plane(horizontal: &Vec<u16>, vertical: &Vec<u16>) -> Option<usize> {
    let cases = [(horizontal, 100, false), (horizontal, 100, true), (vertical, 1, false), (vertical, 1, true)];
    for (grid, mult, invert) in cases.into_iter() {
        let l = grid.len() - 1;
        let mut index = 1;
        while index < grid.len() {
            let mut mirrors = true;
            let mut smudged = false;
            for offset in 0..=(index/2) {
                let abs_diff = if invert {
                    grid[l-offset] ^ grid[l-index+offset]
                } else {
                    grid[offset] ^ grid[index-offset]
                };
                if abs_diff != 0 {
                    if (abs_diff & (abs_diff - 1)) == 0 {
                        if smudged {
                            mirrors = false;
                            break;
                        } else {
                            smudged = true;
                        }
                    } else {
                        mirrors = false;
                        break;
                    }
                }
            }
            if mirrors && smudged {
                if invert {
                    println!("{}", l - (index/2));
                    return Some(mult * (l - (index / 2)));
                } else {
                    println!("{}", (index/2) + 1);
                    return Some(mult * ((index / 2) + 1));
                }
            }
            index += 2;
        }
    }
    return None;
}

pub fn run(input: &Vec<String>) -> AOCResult {
    let mut grid_horizontal = vec![];
    let mut grid_vertical = vec![];
    let mut summary = 0;
    let mut smudged_summary = 0;

    for line in input.into_iter().chain(iter::once(&String::new())) {
        if line == "" {
            summary += find_mirror_plane(&grid_horizontal, &grid_vertical).ok_or("No mirror plane found".to_string())?;
            smudged_summary += find_near_mirror_plane(&grid_horizontal, &grid_vertical).ok_or("No smudged mirror plane found".to_string())?;            
            grid_horizontal = vec![];
            grid_vertical = vec![];
        } else {
            let mut horizontal_number = 0u16;
            grid_vertical.resize(line.len(), 0u16);
            for (byte, vertical_number) in line.as_bytes().into_iter().zip(grid_vertical.iter_mut()) {
                horizontal_number <<= 1;
                *vertical_number <<= 1;
                if *byte == b'#' {
                    horizontal_number |= 1;
                    *vertical_number |= 1;
                }
            }
            grid_horizontal.push(horizontal_number);
        }
    }


    Ok([Some(summary.to_string()), Some(smudged_summary.to_string())])
}

#[test]
pub fn test_day13() {
    assert!(common::run_test(DAY, &run))
}
