use aoc2023_niels_overkamp::common::{self, AOCResult};
use std::mem;
use std::collections::HashMap;
use std::collections::hash_map::Entry;


const DAY: &str = "day14";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
enum Direction {
    North, East, South, West
}

const SPIN_CYCLE: [Direction; 4] = [Direction::North, Direction::West, Direction::South, Direction::East];

fn tilt(grid: &mut Vec<Vec<u8>>, direction: Direction) -> usize {
    let last_x = grid[0].len() - 1;
    let last_y = grid.len() - 1;
    
    let mut load = 0;
    let (last_width, height_iter) = match direction {
        Direction::North => (last_x, (0, last_y, 1)),
        Direction::East => (last_y, (last_x, 0, -1)),
        Direction::South => (last_x, (last_y, 0, -1)),
        Direction::West => (last_y, (0, last_x, 1)),
    };
    let platform_height = grid.len();
    

    for width in 0..=last_width {
//         print!("*{}* ", width);
        let mut height = height_iter.0;
        let mut resting_point = height;
        while height_iter.2 < 0 || height <= height_iter.1 {
            let value = match direction {
                Direction::North | Direction::South => grid[height][width],
                Direction::East | Direction::West => grid[width][height],
            };
//             print!("{} {} ({}), ", resting_point, height, value as char);

            match value {
                b'O' => {
                    if resting_point != height {
//                         print!("[swap {} {}] ", resting_point, height);
                        match direction {
                            Direction::North | Direction::South => { 
                                let (min, max) = (resting_point.min(height), resting_point.max(height));
                                let (left, right) = grid.split_at_mut(max);
                                mem::swap(&mut left[min][width], &mut right[0][width]);
                            }
                            Direction::East | Direction::West => grid[width].swap(resting_point, height),
                        };
                    }
                    load += match direction {
                        Direction::North | Direction::South => platform_height - resting_point,
                        Direction::East | Direction::West => platform_height - width,
                    };
                    if height_iter.2 > 0 {
                        resting_point += 1;
                    } else {
                        resting_point = resting_point.saturating_sub(1);
                    }
                }
                b'#' => {
                    if height_iter.2 > 0 {
                        resting_point = height + 1;
                    } else {
                        resting_point = height.saturating_sub(1);
                    }
                }
                _ => (),
            }
            if height_iter.2 > 0isize {
                height += 1;
            } else if height == 0 {
                break
            } else {
                height -= 1;
            }
        }
//         println!("");
    }
    return load;
}

fn hash_key(grid: &Vec<Vec<u8>>) -> Vec<u128> {
    let mut rows = vec![];
    for row in grid.into_iter() {
        let mut number = 0;
        for byte in row.into_iter() {
            number <<=1;
            if *byte == b'O' {
                number |= 1;
            }
        }
        rows.push(number);
    }
    return rows;
}

fn print(grid: &Vec<Vec<u8>>) {
    let mut s = String::new();
    for line in grid.into_iter() {
        s += String::from_utf8(line.clone()).unwrap().as_str();
        s.push('\n');
    }
    println!("{}", s);
}

pub fn run(input: &Vec<String>) -> AOCResult {
    let mut grid: Vec<Vec<u8>> = input.into_iter().map(|string| Vec::from(string.as_bytes())).collect();
    print(&grid);
    
    let mut visited = HashMap::new();
    let mut load_vec = vec![];
    
    let first_load = tilt(&mut grid, Direction::North);
    visited.insert((Direction::North, hash_key(&grid)), 0);
    load_vec.push(first_load);
    print(&grid);
    
    let mut load = 0;
    
    let n = 4000000000;
    
    
    for (i, direction) in (1..n).into_iter().zip(SPIN_CYCLE.iter().cycle().skip(1)) {
//         println!("{} {:?}", i, direction);
        load = tilt(&mut grid, *direction);
//         print(&grid);
        load_vec.push(load);
        let key = (*direction, hash_key(&grid));
        match visited.entry(key) {
            Entry::Vacant(v) => { v.insert(i); }
            Entry::Occupied(o) => { 
                let cycle_i = o.get();
                let cycle_size = i - cycle_i;
                load = load_vec[cycle_i + ((n - i - 1) % cycle_size)];
                println!("Found loop at {}-{} ({}): {}", cycle_i, i, cycle_size, load);
//                 println!("{:?}", load_vec);
                break;
            }
        }
    }

    Ok([Some(first_load.to_string()), Some(load.to_string())])
}

#[test]
pub fn test_day14() {
    assert!(common::run_test(DAY, &run))
}
