use aoc2023_niels_overkamp::common::{self, AOCResult};
use std::collections::hash_map::{HashMap, Entry};

const DAY: &str = "day15";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

fn HASH_hash(string: &str) -> u8 {
    let mut acc: u8 = 0;
    for byte in string.as_bytes().into_iter() {
        if *byte != b'\n' {
            acc = acc.wrapping_add(*byte);
            acc = acc.wrapping_add(acc << 4); // * 17
        }
    }
    return acc;
}

pub fn run(input: &Vec<String>) -> AOCResult {
    let mut sum: usize = 0;
    
    let mut HASHMAP_vec = vec![HashMap::<_, (usize, u8)>::new(); 255];
    let mut time: usize = 0;
    
    for line in input.into_iter() {
        for instruction in line.split(',') {
            sum += HASH_hash(instruction) as usize; // Part 1
            
            if let Some((label, lens)) = instruction.split_once('=') {
                let lens_box = &mut HASHMAP_vec[HASH_hash(label) as usize];
                let lens_focal = lens.parse::<u8>().unwrap();
                match lens_box.entry(label) {
                    Entry::Occupied(mut e) => { (*e.get_mut()).1 = lens_focal; }
                    Entry::Vacant(mut e) => {
                        e.insert((time, lens_focal));
                        time += 1;
                    }
                }
            } else if let Some(label) = instruction.strip_suffix('-') {
                let lens_box = &mut HASHMAP_vec[HASH_hash(label) as usize];
                lens_box.remove(label);
            }
        }
    }
    
    let mut focusing_power = 0;
    
    for (i, lens_box) in HASHMAP_vec.into_iter().enumerate() {
        let mut lens_box = lens_box.values().collect::<Vec<_>>();
        lens_box.sort();
        for (j, (_, focal)) in lens_box.into_iter().enumerate() {
            focusing_power += (i + 1) * (j + 1) * (*focal as usize);
        }
    }

    Ok([Some(sum.to_string()), Some(focusing_power.to_string())])
}

#[test]
pub fn test_day15() {
    assert!(common::run_test(DAY, &run))
}
