use aoc2023_niels_overkamp::common::{self, AOCResult};
use nom::sequence::{preceded, terminated, tuple};
use nom::bytes::complete::tag;
use nom::character::complete::{u32 as u32_parser, alpha1};
use nom::multi::separated_list0;
use nom::combinator::all_consuming;
use nom::error::{VerboseError, convert_error};
use nom::Finish;
use std::collections::HashMap;

const DAY: &str = "day2";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

pub fn run<'a>(input: &'a Vec<String>) -> AOCResult {
    let total = HashMap::from([("red", 12),("green", 13),("blue", 14)]);
    
    let id_parser = preceded(tag("Game "), terminated(u32_parser::<&'a str, VerboseError<&'a str>>, tag(": ")));
    let cubes_parser = tuple((terminated(u32_parser, tag(" ")), alpha1));
    let pull_parser = separated_list0(tag(", "), cubes_parser);
    let mut game_parser = all_consuming(tuple((id_parser, separated_list0(tag("; "), pull_parser))));
    
    let mut count1 = 0;
    let mut count2 = 0;
    
    for line in input.iter() {
        let output = game_parser(line.as_str())
            .finish()
            .map_err(|e| convert_error(line.as_str(), e))?;
            
        let (_res, (id, game)) = output;
            
        let mut valid = true;
        let mut min_amounts = HashMap::from([("red", 0), ("green", 0), ("blue", 0)]);
        for pull in game {
            let mut amounts = HashMap::from([("red", 0), ("green", 0), ("blue", 0)]);
            for (amount, color) in pull {
                if let Some(v) = amounts.get_mut(color) {
                    *v += amount;
                }
                if let Some(v) = min_amounts.get_mut(color) {
                    *v = amount.max(*v);
                }
            }
            if valid {
                for color in ["red", "green", "blue"] {
                    if total[color] < amounts[color] {
                        valid = false;
                    }
                }
            }
        }
        if valid {
            count1 += id;
        }
        count2 += min_amounts.into_values().product::<u32>();
    }

    Ok([Some(count1.to_string()), Some(count2.to_string())])
}

#[test]
pub fn test_day2() {
    assert!(common::run_test(DAY, &run))
}
