use aoc2023_niels_overkamp::common::{self, AOCResult};
use std::collections::{HashSet, HashMap};

const DAY: &str = "day3";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

struct Code {
    start_x: usize,
    end_x: usize,
    y: usize,
    code: usize
}



pub fn run(input: &Vec<String>) -> AOCResult {
    let mut symbols = HashSet::new();
    let mut gears = HashMap::new();
    
/*    let connects_le = |x, y, ss: &mut HashSet<_>, sg: &mut HashMap<_, _>| {
        if x > 0 {
            let symbol_connects = ss.contains(&(x - 1, y)) 
                                || ss.contains(&(x - 1, y + 1))
                                || (y > 0 && ss.contains(&(x - 1, y - 1)));
            
            let connecting_gear = sg.get_mut(&(x - 1, y)) 
                .or(sg.get_mut(&(x - 1, y + 1)))
                .or(sg.get_mut(&(x - 1, y - 1)));
                
            (symbol_connects, connecting_gear)
        } else {
            let symbol_connects = ss.contains(&(x, y + 1)) 
                                || (y > 0 && ss.contains(&(x, y - 1)));
                                
            let connecting_gear = sg.get_mut(&(x, y + 1)) 
                .or(sg.get_mut(&(x, y - 1)));
            
            (symbol_connects, connecting_gear)
        }
    };
    
    let connects_gt = |x, y, ss: &mut HashSet<_>, sg: &mut HashMap<_, _>| {
            let symbol_connects = ss.contains(&(x, y)) 
                                || ss.contains(&(x, y + 1)) 
                                || (y > 0 && ss.contains(&(x, y - 1)));
                                
            let connecting_gear = sg.get_mut(&(x, y))
                .or(sg.get_mut(&(x, y + 1))) 
                .or(sg.get_mut(&(x, y - 1)));
            
            (symbol_connects, connecting_gear)
    };
 */   
    for (y, row) in input.iter().enumerate() {
        for (x, c) in row.chars().enumerate() {
            if !(c.is_digit(10) || c == '.') {
                symbols.insert((x,y));
            }
            if c == '*' {
                gears.insert((x,y), [1, 0]);
            }
        }
    }
    
    let mut codes = Vec::new();    
    
    for (y, row) in input.iter().enumerate() {
        let mut code = None;
        for (x, c) in row.chars().enumerate() {
            if let Some(d) = c.to_digit(10) {
                let code = code.get_or_insert(Code {start_x: x, end_x: x, y: y, code: 0});
                code.code *= 10;
                code.code += d as usize;
                code.end_x = x;
            } else if let Some(code) = code.take() {
                codes.push(code);
            }
        }
        if let Some(code) = code {
            codes.push(code);
        }
    }
    
    let mut sum = 0;
    for Code { start_x, end_x, y, code } in codes {
        let mut connected = false;
        for x in (1.max(start_x) - 1)..=(end_x+1) {
            for y in (1.max(y) - 1)..=(y+1) {
                if !connected && symbols.contains(&(x,y)) {
                    connected = true;
                    sum += code;
                }
                if let Some(gear) = gears.get_mut(&(x,y)) {
                    let [ratio, count] = *gear;
                    if count == 1 {
                        *gear = [ratio * code, 2]
                    } else if count == 0 {
                        *gear = [code, 1]
                    } else {
                        *gear = [0, count + 1]
                    }
                }
            }
        }
    }
    
    let mut gear_sum = 0;
    for [ratio, count] in gears.values() {
        if *count == 2 {
            gear_sum += ratio;
        }
    }
    
    Ok([Some(sum.to_string()), Some(gear_sum.to_string())])
}

#[test]
pub fn test_day3() {
    assert!(common::run_test(DAY, &run))
}
