use aoc2023_niels_overkamp::common::{self, AOCResult};
use nom::combinator::all_consuming;
use nom::sequence::{preceded, tuple};
use nom::bytes::complete::{tag, take};
use nom::character::complete::{u64 as u64_parser, alpha1, multispace1};
use nom::multi::{separated_list0, fold_many0};
use nom::error::convert_error;
use nom::Finish;

const DAY: &str = "day6";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

pub fn run(input: &Vec<String>) -> AOCResult {
    let mut numbers_parser = all_consuming(preceded(tuple((alpha1, tag(":"), multispace1)), separated_list0(multispace1, u64_parser)));
    let mut numbers_parser = |line| numbers_parser(line)
        .finish()
        .map_err(|e| convert_error(line, e));
        
    let mut number_parser = all_consuming(preceded(
        tuple((alpha1, tag(":"), multispace1)), 
        fold_many0(take(1usize), || 0, |acc: u64, digit: &str| if digit != " " { 
                digit.parse::<u64>().unwrap() + acc * 10
            } else {
                acc
            })));
    let mut number_parser = |line| number_parser(line)
        .finish()
        .map_err(|e| convert_error(line, e));

    let (_, times) = numbers_parser(input[0].as_str()).unwrap();
    let (_, distances) = numbers_parser(input[1].as_str()).unwrap();
    
    let (_, time) = number_parser(input[0].as_str()).unwrap();
    let (_, distance) = number_parser(input[1].as_str()).unwrap();
    
    let calc_margin = |time, distance| {
        let time = time as f64;
        let distance = distance as f64;
        let square_root_discriminator = (time*time - 4f64 * distance).sqrt();
        let lower_bound = ((time - square_root_discriminator) / 2f64) + 10f64 * f64::EPSILON;
        let upper_bound = ((time + square_root_discriminator) / 2f64) - 10f64 * f64::EPSILON;
        (upper_bound.floor() - lower_bound.ceil() + 1f64) as u64
    };
    
    let mut mul = 1;
    for (time, distance) in times.into_iter().zip(distances.into_iter()) {
        let margin = calc_margin(time, distance);
        println!("{}", margin);
        mul *= margin;
    }
    
    Ok([Some(mul.to_string()), Some(calc_margin(time, distance).to_string())])
}

#[test]
pub fn test_day6() {
    assert!(common::run_test(DAY, &run))
}
